@@ -1,5 +1,5 @@
 /* ----------------------------------------------------------------------- *
- *   
+ *
  *   Copyright 1996-2016 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
@@ -14,7 +14,7 @@
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
- *     
+ *
  *     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  *     CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  *     INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
@@ -67,12 +67,9 @@
  */
 #include "compiler.h"
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
 #include <time.h>
-#include <stdarg.h>             /* Note: we need the ANSI version of stdarg.h */
-#include <ctype.h>
+#include <ctype.h>              /* For toupper() */
+#include "nctype.h"
 
 #include "nasm.h"
 #include "nasmlib.h"
@@ -146,11 +143,12 @@ static struct ieeeSection {
     struct ieeeObjData *data, *datacurr;
     struct ieeeFixupp *fptr, *flptr;
     int32_t index;                 /* the NASM segment id */
-    int32_t ieee_index;            /* the OBJ-file segment index */
+    int32_t ieee_index;            /* the IEEE-file segment index */
     int32_t currentpos;
     int32_t align;                 /* can be SEG_ABS + absolute addr */
     int32_t startpos;
     int32_t use32;                 /* is this segment 32-bit? */
+    int64_t pass_last_seen;
     struct ieeePublic *pubhead, **pubtail, *lochead, **loctail;
     enum {
         CMB_PRIVATE = 0,
@@ -193,7 +191,7 @@ static void ieee_data_new(struct ieeeSection *);
 static void ieee_write_fixup(int32_t, int32_t, struct ieeeSection *,
                              int, uint64_t, int32_t);
 static void ieee_install_fixup(struct ieeeSection *, struct ieeeFixupp *);
-static int32_t ieee_segment(char *, int, int *);
+static int32_t ieee_segment(char *, int *);
 static void ieee_write_file(void);
 static void ieee_write_byte(struct ieeeSection *, int);
 static void ieee_write_word(struct ieeeSection *, int);
@@ -209,6 +207,7 @@ static void ieee_unqualified_name(char *, char *);
  */
 static void ieee_init(void)
 {
+    strlcpy(ieee_infile, inname, sizeof(ieee_infile));
     any_segs = false;
     fpubhead = NULL;
     fpubtail = &fpubhead;
@@ -293,9 +292,8 @@ static void ieee_deflabel(char *name, int32_t segment,
     struct ieeeSection *seg;
     int i;
 
-    if (special) {
-        nasm_error(ERR_NONFATAL, "unrecognised symbol type `%s'", special);
-    }
+    if (special)
+        nasm_nonfatal("unrecognised symbol type `%s'", special);
     /*
      * First check for the double-period, signifying something
      * unusual.
@@ -397,24 +395,14 @@ static void ieee_out(int32_t segto, const void *data,
     int32_t ldata;
     struct ieeeSection *seg;
 
-    /*
-     * handle absolute-assembly (structure definitions)
-     */
-    if (segto == NO_SEG) {
-        if (type != OUT_RESERVE)
-            nasm_error(ERR_NONFATAL, "attempt to assemble code in [ABSOLUTE]"
-                  " space");
-        return;
-    }
-
     /*
      * If `any_segs' is still false, we must define a default
      * segment.
      */
     if (!any_segs) {
         int tempint;            /* ignored */
-        if (segto != ieee_segment("__NASMDEFSEG", 2, &tempint))
-            nasm_panic(0, "strange segment conditions in IEEE driver");
+        if (segto != ieee_segment("__NASMDEFSEG", &tempint))
+            nasm_panic("strange segment conditions in IEEE driver");
     }
 
     /*
@@ -424,7 +412,7 @@ static void ieee_out(int32_t segto, const void *data,
         if (seg->index == segto)
             break;
     if (!seg)
-        nasm_panic(0, "code directed to nonexistent segment?");
+        nasm_panic("code directed to nonexistent segment?");
 
     if (type == OUT_RAWDATA) {
         ucdata = data;
@@ -435,8 +423,8 @@ static void ieee_out(int32_t segto, const void *data,
         if (type == OUT_ADDRESS)
             size = abs((int)size);
         else if (segment == NO_SEG)
-            nasm_error(ERR_NONFATAL, "relative call to absolute address not"
-                  " supported by IEEE format");
+            nasm_nonfatal("relative call to absolute address not"
+                          " supported by IEEE format");
         ldata = *(int64_t *)data;
         if (type == OUT_REL2ADR)
             ldata += (size - 2);
@@ -525,15 +513,13 @@ static void ieee_write_fixup(int32_t segment, int32_t wrt,
                             s.addend = 0;
                             s.id2 = eb->index[i];
                         } else
-                            nasm_error(ERR_NONFATAL,
-                                  "Source of WRT must be an offset");
+                            nasm_nonfatal("source of WRT must be an offset");
                     }
 
                 } else
-                    nasm_panic(0,
-                          "unrecognised WRT value in ieee_write_fixup");
+                    nasm_panic("unrecognised WRT value in ieee_write_fixup");
             } else
-                nasm_error(ERR_NONFATAL, "target of WRT must be a section ");
+                nasm_nonfatal("target of WRT must be a section");
         }
         s.size = size;
         ieee_install_fixup(segto, &s);
@@ -572,8 +558,7 @@ static void ieee_write_fixup(int32_t segment, int32_t wrt,
                  */
                 if (eb) {
                     if (realtype == OUT_REL2ADR || realtype == OUT_REL4ADR) {
-                        nasm_panic(0,
-                              "Segment of a rel not supported in ieee_write_fixup");
+                        nasm_panic("Segment of a rel not supported in ieee_write_fixup");
                     } else {
                         /* If we want the segment */
                         s.ftype = FT_EXTSEG;
@@ -583,8 +568,7 @@ static void ieee_write_fixup(int32_t segment, int32_t wrt,
 
                 } else
                     /* If we get here the seg value doesn't make sense */
-                    nasm_panic(0,
-                          "unrecognised segment value in ieee_write_fixup");
+                    nasm_panic("unrecognised segment value in ieee_write_fixup");
             }
 
         } else {
@@ -638,13 +622,12 @@ static void ieee_write_fixup(int32_t segment, int32_t wrt,
 
                 } else
                     /* If we get here the seg value doesn't make sense */
-                    nasm_panic(0,
-                          "unrecognised segment value in ieee_write_fixup");
+                    nasm_panic("unrecognised segment value in ieee_write_fixup");
             }
         }
         if (size != 2 && s.ftype == FT_SEG)
-            nasm_error(ERR_NONFATAL, "IEEE format can only handle 2-byte"
-                  " segment base references");
+            nasm_nonfatal("IEEE format can only handle 2-byte"
+                          " segment base references");
         s.size = size;
         ieee_install_fixup(segto, &s);
         return;
@@ -670,7 +653,7 @@ static void ieee_install_fixup(struct ieeeSection *seg,
 /*
  * segment registry
  */
-static int32_t ieee_segment(char *name, int pass, int *bits)
+static int32_t ieee_segment(char *name, int *bits)
 {
     /*
      * We call the label manager here to define a name for the new
@@ -720,13 +703,15 @@ static int32_t ieee_segment(char *name, int pass, int *bits)
         for (seg = seghead; seg; seg = seg->next) {
             ieee_idx++;
             if (!strcmp(seg->name, name)) {
-                if (attrs > 0 && pass == 1)
-                    nasm_error(ERR_WARNING, "segment attributes specified on"
-                          " redeclaration of segment: ignoring");
+                if (attrs > 0 && seg->pass_last_seen == pass_count())
+                    nasm_warn(WARN_OTHER, "segment attributes specified on"
+                              " redeclaration of segment: ignoring");
                 if (seg->use32)
                     *bits = 32;
                 else
                     *bits = 16;
+
+                seg->pass_last_seen = pass_count();
                 return seg->index;
             }
         }
@@ -777,8 +762,7 @@ static int32_t ieee_segment(char *name, int pass, int *bits)
                     seg->align = 1;
                 if (rn_error) {
                     seg->align = 1;
-                    nasm_error(ERR_NONFATAL, "segment alignment should be"
-                          " numeric");
+                    nasm_nonfatal("segment alignment should be numeric");
                 }
                 switch (seg->align) {
                 case 1:        /* BYTE */
@@ -792,25 +776,23 @@ static int32_t ieee_segment(char *name, int pass, int *bits)
                 case 128:
                     break;
                 default:
-                    nasm_error(ERR_NONFATAL, "invalid alignment value %d",
-                          seg->align);
+                    nasm_nonfatal("invalid alignment value %d", seg->align);
                     seg->align = 1;
                     break;
                 }
             } else if (!nasm_strnicmp(p, "absolute=", 9)) {
                 seg->align = SEG_ABS + readnum(p + 9, &rn_error);
                 if (rn_error)
-                    nasm_error(ERR_NONFATAL, "argument to `absolute' segment"
-                          " attribute should be numeric");
+                    nasm_nonfatal("argument to `absolute' segment"
+                                  " attribute should be numeric");
             }
         }
 
         ieee_seg_needs_update = seg;
         if (seg->align >= SEG_ABS)
-            define_label(name, NO_SEG, seg->align - SEG_ABS,
-			 NULL, false, false);
+            define_label(name, NO_SEG, seg->align - SEG_ABS, false);
         else
-            define_label(name, seg->index + 1, 0L, NULL, false, false);
+            define_label(name, seg->index + 1, 0L, false);
         ieee_seg_needs_update = NULL;
 
         if (seg->use32)
@@ -825,17 +807,15 @@ static int32_t ieee_segment(char *name, int pass, int *bits)
  * directives supported
  */
 static enum directive_result
-ieee_directive(enum directive directive, char *value, int pass)
+ieee_directive(enum directive directive, char *value)
 {
-
     (void)value;
-    (void)pass;
 
     switch (directive) {
     case D_UPPERCASE:
         ieee_uppercase = true;
         return DIRR_OK;
-	
+
     default:
 	return DIRR_UNKNOWN;
     }
@@ -886,19 +866,9 @@ static int32_t ieee_segbase(int32_t segment)
     return segment;             /* no special treatment */
 }
 
-/*
- * filename
- */
-static void ieee_filename(char *inname, char *outname)
-{
-    strcpy(ieee_infile, inname);
-    standard_extension(inname, outname, ".o");
-}
-
 static void ieee_write_file(void)
 {
-    struct tm *thetime;
-    time_t reltime;
+    const struct tm * const thetime = &official_compile_time.local;
     struct FileName *fn;
     struct ieeeSection *seg;
     struct ieeePublic *pub, *loc;
@@ -917,7 +887,7 @@ static void ieee_write_file(void)
     /*
      * Write the NASM boast comment.
      */
-    ieee_putascii("CO0,%02X%s.\n", strlen(nasm_comment), nasm_comment);
+    ieee_putascii("CO0,%02X%s.\n", nasm_comment_len(), nasm_comment());
 
     /*
      * write processor-specific information
@@ -927,8 +897,6 @@ static void ieee_write_file(void)
     /*
      * date and time
      */
-    time(&reltime);
-    thetime = localtime(&reltime);
     ieee_putascii("DT%04d%02d%02d%02d%02d%02d.\n",
                   1900 + thetime->tm_year, thetime->tm_mon + 1,
                   thetime->tm_mday, thetime->tm_hour, thetime->tm_min,
@@ -991,7 +959,7 @@ static void ieee_write_file(void)
             if (seg->index == ieee_entry_seg)
                 break;
         if (!seg)
-            nasm_panic(0, "Start address records are incorrect");
+            nasm_panic("Start address records are incorrect");
         else
             ieee_putascii("ASG,R%X,%lX,+.\n", seg->ieee_index,
                           ieee_entry_ofs);
@@ -1305,7 +1273,7 @@ static void dbgls_init(void)
     arrindex = ARRAY_BOT;
     arrhead = NULL;
     arrtail = &arrhead;
-    ieee_segment("??LINE", 2, &tempint);
+    ieee_segment("??LINE", &tempint);
     any_segs = false;
 }
 static void dbgls_cleanup(void)
@@ -1353,8 +1321,8 @@ static void dbgls_linnum(const char *lnfname, int32_t lineno, int32_t segto)
      */
     if (!any_segs) {
         int tempint;            /* ignored */
-        if (segto != ieee_segment("__NASMDEFSEG", 2, &tempint))
-            nasm_panic(0, "strange segment conditions in OBJ driver");
+        if (segto != ieee_segment("__NASMDEFSEG", &tempint))
+            nasm_panic("strange segment conditions in IEEE driver");
     }
 
     /*
@@ -1364,7 +1332,7 @@ static void dbgls_linnum(const char *lnfname, int32_t lineno, int32_t segto)
         if (seg->index == segto)
             break;
     if (!seg)
-        nasm_panic(0, "lineno directed to nonexistent segment?");
+        nasm_panic("lineno directed to nonexistent segment?");
 
     for (fn = fnhead; fn; fn = fn->next) {
         if (!nasm_stricmp(lnfname, fn->name))
@@ -1510,20 +1478,22 @@ static const struct dfmt * const ladsoft_debug_arr[3] = {
 const struct ofmt of_ieee = {
     "IEEE-695 (LADsoft variant) object file format",
     "ieee",
+    ".o",
     OFMT_TEXT,
     32,
     ladsoft_debug_arr,
     &ladsoft_debug_form,
     NULL,
     ieee_init,
+    null_reset,
     nasm_do_legacy_output,
     ieee_out,
     ieee_deflabel,
     ieee_segment,
+    NULL,
     ieee_sectalign,
     ieee_segbase,
     ieee_directive,
-    ieee_filename,
     ieee_cleanup,
     NULL                        /* pragma list */
 };
