@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
- *   
- *   Copyright 1996-2009 The NASM Authors - All Rights Reserved
+ *
+ *   Copyright 1996-2014 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -14,7 +14,7 @@
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
- *     
+ *
  *     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  *     CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  *     INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
@@ -54,18 +54,15 @@
 
 #include <stdio.h>
 #include <stdlib.h>
-#include <string.h>
 
-#define RDOFF_UTILS
-
-#include "rdoff.h"
+#include "rdfutils.h"
 #include "symtab.h"
 #include "collectn.h"
 #include "rdlib.h"
 #include "segtab.h"
 #include "nasmlib.h"
 
-#define LDRDF_VERSION "1.07"
+#define LDRDF_VERSION "1.08"
 
 /* #define STINGY_MEMORY */
 
@@ -91,9 +88,6 @@ struct modulenode {
 
 #include "ldsegs.h"
 
-#define newstr(str) strcpy(malloc(strlen(str) + 1),str)
-#define newstrcat(s1,s2) strcat(strcpy(malloc(strlen(s1)+strlen(s2)+1),s1),s2)
-
 /* ==========================================================================
  * Function prototypes of private utility functions
  */
@@ -128,8 +122,8 @@ char *libpath = NULL;
 /* file to embed as a generic record */
 char *generic_rec_file = NULL;
 
-/* error file */
-static FILE *error_file;
+/* module name to be added at the beginning of output file */
+char *modname_specified = NULL;
 
 /* the header of the output file, built up stage by stage */
 rdf_headerbuf *newheader = NULL;
@@ -165,7 +159,7 @@ int errorcount = 0;             /* determines main program exit status */
  *
  * sets up segments 0, 1, and 2, the initial code data and bss segments
  */
-void initsegments()
+static void initsegments(void)
 {
     nsegs = 3;
     outputseg[0].type = 1;
@@ -184,23 +178,23 @@ void initsegments()
 }
 
 /*
- * loadmodule
+ * loadmodule()
  *
  * Determine the characteristics of a module, and decide what to do with
  * each segment it contains (including determining destination segments and
  * relocation factors for segments that	are kept).
  */
-void loadmodule(const char *filename)
+static void loadmodule(const char *filename)
 {
     if (options.verbose)
         printf("loading `%s'\n", filename);
 
     /* allocate a new module entry on the end of the modules list */
     if (!modules) {
-        modules = malloc(sizeof(*modules));
+        modules = nasm_malloc(sizeof(*modules));
         lastmodule = modules;
     } else {
-        lastmodule->next = malloc(sizeof(*modules));
+        lastmodule->next = nasm_malloc(sizeof(*modules));
         lastmodule = lastmodule->next;
     }
 
@@ -221,7 +215,7 @@ void loadmodule(const char *filename)
      * factor if we decide to keep them)
      */
     lastmodule->header = NULL;
-    lastmodule->name = strdup(filename);
+    lastmodule->name = nasm_strdup(filename);
     lastmodule->next = NULL;
 
     processmodule(filename, lastmodule);
@@ -322,7 +316,7 @@ void processmodule(const char *filename, struct modulenode *mod)
      * extract symbols from the header, and dump them into the
      * symbol table
      */
-    header = malloc(mod->f.header_len);
+    header = nasm_malloc(mod->f.header_len);
     if (!header) {
         fprintf(stderr, "ldrdf: not enough memory\n");
         exit(1);
@@ -417,7 +411,7 @@ void processmodule(const char *filename, struct modulenode *mod)
      * of this program...
      */
     mod->f.header_loc = NULL;
-    free(header);
+    nasm_free(header);
 
 #endif
 
@@ -426,7 +420,7 @@ void processmodule(const char *filename, struct modulenode *mod)
 /*
  * Return 1 if a given module is in the list, 0 otherwise.
  */
-int lookformodule(const char *name)
+static int lookformodule(const char *name)
 {
     struct modulenode *curr = modules;
 
@@ -521,12 +515,12 @@ void symtab_add(const char *symbol, int segment, int32_t offset)
     /*
      * this is the first declaration of this symbol
      */
-    ste = malloc(sizeof(symtabEnt));
+    ste = nasm_malloc(sizeof(symtabEnt));
     if (!ste) {
         fprintf(stderr, "ldrdf: out of memory\n");
         exit(1);
     }
-    ste->name = strdup(symbol);
+    ste->name = nasm_strdup(symbol);
     ste->segment = segment;
     ste->offset = offset;
     ste->flags = 0;
@@ -560,7 +554,7 @@ int symtab_get(const char *symbol, int *segment, int32_t *offset)
  * checks that a library can be opened and is in the correct format,
  * then adds it to the linked list of libraries.
  */
-void add_library(const char *name)
+static void add_library(const char *name)
 {
     if (rdl_verify(name)) {
         rdl_perror("ldrdf", name);
@@ -568,13 +562,13 @@ void add_library(const char *name)
         return;
     }
     if (!libraries) {
-        lastlib = libraries = malloc(sizeof(*libraries));
+        lastlib = libraries = nasm_malloc(sizeof(*libraries));
         if (!libraries) {
             fprintf(stderr, "ldrdf: out of memory\n");
             exit(1);
         }
     } else {
-        lastlib->next = malloc(sizeof(*libraries));
+        lastlib->next = nasm_malloc(sizeof(*libraries));
         if (!lastlib->next) {
             fprintf(stderr, "ldrdf: out of memory\n");
             exit(1);
@@ -599,7 +593,7 @@ void add_library(const char *name)
  * returns 1 if any extra library modules are included, indicating that
  * another pass through the library list should be made (possibly).
  */
-int search_libraries()
+static int search_libraries(void)
 {
     struct librarynode *cur;
     rdffile f;
@@ -623,7 +617,7 @@ int search_libraries()
             if (options.verbose > 3)
                 printf("  looking in module `%s'\n", f.name);
 
-            header = malloc(f.header_len);
+            header = nasm_malloc(f.header_len);
             if (!header) {
                 fprintf(stderr, "ldrdf: not enough memory\n");
                 exit(1);
@@ -668,20 +662,20 @@ int search_libraries()
                  * there are modules on the module list by the time
                  * we get here.
                  */
-                lastmodule->next = malloc(sizeof(*lastmodule->next));
+                lastmodule->next = nasm_malloc(sizeof(*lastmodule->next));
                 if (!lastmodule->next) {
                     fprintf(stderr, "ldrdf: not enough memory\n");
                     exit(1);
                 }
                 lastmodule = lastmodule->next;
                 memcpy(&lastmodule->f, &f, sizeof(f));
-                lastmodule->name = strdup(f.name);
+                lastmodule->name = nasm_strdup(f.name);
                 lastmodule->next = NULL;
                 processmodule(f.name, lastmodule);
                 break;
             }
             if (!keepfile) {
-                free(f.name);
+                nasm_free(f.name);
                 f.name = NULL;
                 f.fp = NULL;
             }
@@ -706,12 +700,12 @@ int search_libraries()
  * all the modules into a single output module, and then writes this to a
  * file.
  */
-void write_output(const char *filename)
+static void write_output(const char *filename)
 {
     FILE *f;
     rdf_headerbuf *rdfheader;
     struct modulenode *cur;
-    int i, availableseg, seg, localseg, isrelative;
+    int i, n, availableseg, seg, localseg, isrelative;
     void *header;
     rdfheaderrec *hr, newrec;
     symtabEnt *se;
@@ -741,13 +735,13 @@ void write_output(const char *filename)
             printf("\nadding generic record from binary file %s\n",
                    generic_rec_file);
 
-        hr = (rdfheaderrec *) malloc(sizeof(struct GenericRec));
+        hr = (rdfheaderrec *) nasm_malloc(sizeof(struct GenericRec));
         if ((ff = fopen(generic_rec_file, "r")) == NULL) {
             fprintf(stderr, "ldrdf: couldn't open %s for input\n",
                     generic_rec_file);
             exit(1);
         }
-        i = fread(hr->g.data, 1, sizeof(hr->g.data), ff);
+        n = fread(hr->g.data, 1, sizeof(hr->g.data), ff);
         fseek(ff, 0, SEEK_END);
         if (ftell(ff) > (long)sizeof(hr->g.data)) {
             fprintf(error_file,
@@ -757,12 +751,36 @@ void write_output(const char *filename)
         }
         fclose(ff);
 
-        hr->g.type = 0;
-        hr->g.reclen = i;
+        hr->g.type = RDFREC_GENERIC;
+        hr->g.reclen = n;
         rdfaddheader(rdfheader, hr);
-        free(hr);
+        nasm_free(hr);
     }
 
+    /*
+     * Add module name record if `-mn' option was given
+     */
+    if (modname_specified) {
+	n = strlen(modname_specified);
+
+	if ((n < 1) || (n >= MODLIB_NAME_MAX)) {
+            fprintf(stderr, "ldrdf: invalid length of module name `%s'\n",
+		modname_specified);
+            exit(1);
+        }
+
+        if (options.verbose)
+            printf("\nadding module name record %s\n", modname_specified);
+
+        hr = (rdfheaderrec *) nasm_malloc(sizeof(struct ModRec));
+	hr->m.type = RDFREC_MODNAME;
+        hr->m.reclen = n + 1;
+        strcpy(hr->m.modname, modname_specified);
+        rdfaddheader(rdfheader, hr);
+        nasm_free(hr);
+    }
+
+
     if (options.verbose)
         printf("\nbuilding output module (%d segments)\n", nsegs);
 
@@ -776,7 +794,7 @@ void write_output(const char *filename)
         outputseg[i].data = NULL;
         if (!outputseg[i].length)
             continue;
-        outputseg[i].data = malloc(outputseg[i].length);
+        outputseg[i].data = nasm_malloc(outputseg[i].length);
         if (!outputseg[i].data) {
             fprintf(stderr, "ldrdf: out of memory\n");
             exit(1);
@@ -814,7 +832,7 @@ void write_output(const char *filename)
          * Perform fixups, and add new header records where required
          */
 
-        header = malloc(cur->f.header_len);
+        header = nasm_malloc(cur->f.header_len);
         if (!header) {
             fprintf(stderr, "ldrdf: out of memory\n");
             exit(1);
@@ -973,12 +991,12 @@ void write_output(const char *filename)
                      * future reference
                      */
                     if (!se) {
-                        se = malloc(sizeof(*se));
+                        se = nasm_malloc(sizeof(*se));
                         if (!se) {
                             fprintf(stderr, "ldrdf: out of memory\n");
                             exit(1);
                         }
-                        se->name = strdup(hr->i.label);
+                        se->name = nasm_strdup(hr->i.label);
                         se->flags = 0;
                         se->segment = availableseg++;
                         se->offset = 0;
@@ -1098,7 +1116,7 @@ void write_output(const char *filename)
             }
         }
 
-        free(header);
+        nasm_free(header);
         done_seglocations(&segs);
 
     }
@@ -1128,22 +1146,14 @@ void write_output(const char *filename)
      * the output file
      */
     for (i = 0; i < nsegs; i++) {
-        uint16_t s;
-        int32_t l;
-
         if (i == 2)
             continue;
 
-        s = translateint16_t(outputseg[i].type);
-        fwrite(&s, 2, 1, f);
-        s = translateint16_t(outputseg[i].number);
-        fwrite(&s, 2, 1, f);
-        s = translateint16_t(outputseg[i].reserved);
-        fwrite(&s, 2, 1, f);
-        l = translateint32_t(outputseg[i].length);
-        fwrite(&l, 4, 1, f);
-
-        fwrite(outputseg[i].data, outputseg[i].length, 1, f);
+        fwriteint16_t(outputseg[i].type, f);
+        fwriteint16_t(outputseg[i].number, f);
+        fwriteint16_t(outputseg[i].reserved, f);
+        fwriteint32_t(outputseg[i].length, f);
+        nasm_write(outputseg[i].data, outputseg[i].length, f);
     }
 
     fwritezero(10, f);
@@ -1153,7 +1163,7 @@ void write_output(const char *filename)
  * Main program
  */
 
-void usage()
+static void usage(void)
 {
     printf("usage:\n"
            "   ldrdf [options] object modules ... [-llibrary ...]\n"
@@ -1166,7 +1176,8 @@ void usage()
            "   -o name         write output in file 'name'\n"
            "   -j path         specify objects search path\n"
            "   -L path         specify libraries search path\n"
-           "   -g file         embed 'file' as a first header record with type 'generic'\n");
+           "   -g file         embed 'file' as a first header record with type 'generic'\n"
+           "   -mn name        add module name record at the beginning of output file\n");
     exit(0);
 }
 
@@ -1176,6 +1187,8 @@ int main(int argc, char **argv)
     int moduleloaded = 0;
     char *respstrings[128] = { 0, };
 
+    rdoff_init();
+
     options.verbose = 0;
     options.align = 16;
     options.dynalink = 0;
@@ -1221,6 +1234,16 @@ int main(int argc, char **argv)
             if (argv[0][2] == 'y')
                 options.dynalink = 1;
             break;
+        case 'm':
+            if (argv[0][2] == 'n') {
+                modname_specified = argv[1];
+		argv++, argc--;
+		if (!argc) {
+		    fprintf(stderr, "ldrdf: -mn expects a module name\n");
+		    exit(1);
+		}
+	    }
+            break;
         case 'o':
             outname = argv[1];
             argv++, argc--;
@@ -1279,7 +1302,7 @@ int main(int argc, char **argv)
                         fprintf(stderr, "ldrdf: too many input files\n");
                         exit(1);
                     }
-                    *(respstrings + i) = newstr(buf);
+                    *(respstrings + i) = nasm_strdup(buf);
                     argc++, i++;
                 }
                 fclose(f);
@@ -1292,6 +1315,10 @@ int main(int argc, char **argv)
         case 'g':
             generic_rec_file = argv[1];
             argv++, argc--;
+            if (!argc) {
+		fprintf(stderr, "ldrdf: -g expects a file name\n");
+		exit(1);
+	    }
             break;
         default:
             usage();
@@ -1329,12 +1356,12 @@ int main(int argc, char **argv)
             break;
         if (!strncmp(*argv, "-l", 2)) {
             if (libpath && (argv[0][2] != '/'))
-                add_library(newstrcat(libpath, *argv + 2));
+                add_library(nasm_strcat(libpath, *argv + 2));
             else
                 add_library(*argv + 2);
         } else {
             if (objpath && (argv[0][0] != '/'))
-                loadmodule(newstrcat(objpath, *argv));
+                loadmodule(nasm_strcat(objpath, *argv));
             else
                 loadmodule(*argv);
             moduleloaded = 1;
@@ -1356,7 +1383,9 @@ int main(int argc, char **argv)
 
     write_output(outname);
 
-    if (errorcount > 0)
+    if (errorcount > 0) {
+        remove(outname);
         exit(1);
+    }
     return 0;
 }
